#! /usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import matplotlib.pyplot as mpl
import warnings

from propobject import BaseObject

from modefit.baseobjects import BaseFitter

from .model import read_psfmodel
from .chromatic_model import stddev_chromaticity
from .tools import kwargs_update
from .chromatic_model import LBDAREF

FITKEY  = "slpsf"
USE_LEASTSQ = True


def fit_slice(slice_, fitbuffer=None,
              psfmodel="NormalMoffatTilted", fitted_indexes=None,
              lbda=None, centroids=None, centroids_err=[2,2],
              adjust_errors=True, force_centroid=False,
              **kwargs):
    """ Fit PSF Slice without forcing it's shape

    Parameters
    ----------

    Returns
    -------
    SlicePSF
    """
    from .tools import kwargs_update
    slpsf = SlicePSF(slice_, psfmodel=psfmodel,
                    fitbuffer=fitbuffer, fitted_indexes=fitted_indexes)

    
    if centroids is None:
        xcentroid, ycentroid = None, None
    elif len(centroids) !=2:
        raise TypeError("given centroid should be None or [x,y]")
    else:
        xcentroid, ycentroid = centroids

    # - Fitting 
    fit_default = slpsf.get_guesses(xcentroid=xcentroid,            ycentroid=ycentroid,
                                    xcentroid_err=centroids_err[0], ycentroid_err=centroids_err[1])
    if force_centroid:
        fit_default["xcentroid_fixed"] = True
        fit_default["ycentroid_fixed"] = True

    fit_parameters = kwargs_update(fit_default, **kwargs)
    slpsf.fit( **fit_parameters )
    # - Fitting
    
    dof = slpsf.npoints - slpsf.model.nparam
    if slpsf.fitvalues["chi2"] / dof>2 and adjust_errors and not USE_LEASTSQ:
        from .tools import fit_intrinsic
        model = slpsf.model.get_model(slpsf._xfitted, slpsf._yfitted)
        intrinsic = fit_intrinsic(slpsf._datafitted, model, slpsf._errorfitted, dof, intrinsic_guess=None)
        slpsf.set_intrinsic_error(intrinsic / np.sqrt(2) )
        slpsf.fit( **fit_parameters )
    
    return slpsf

# ====================== #
#                        #
#    PSF Classes         #
#                        #
# ====================== #
class SlicePSFCollection( BaseObject ):
    """ """
    PROPERTIES = ["slices","cube"]
    DERIVED_PROPERTIES = ["adrfitter"]


    def __init__(self, cube=None):
        """ """
        if cube is not None:
            self.set_cube(cube)
            
    # =================== #
    #   Methods           #
    # =================== #
    # --------- #
    #  GETTER   # 
    # --------- #

    # = Models
    def get_chromatic_profile_model(self):
        """ """
        from . import chromatic_model
        cmodel = chromatic_model.ChromaticNormalMoffat()
        used_slindexes  =  self.slindexes[~self.fetch_outlier()]
        cmodel.set_data( *[self.get_fitted_value(k, slindexes=used_slindexes)
                               for k in ["stddev","alpha","amplitude_ratio","lbda",
                                          "stddev.err","alpha.err","amplitude_ratio.err"] ] )
        
        adrmodel = chromatic_model.ADRModel(self.adrfitter.model.adr, 
                                            self.adrfitter.fitvalues["xref"], self.adrfitter.fitvalues["yref"], 
                                            unit=self.adrfitter.model._unit
                                            )
        cmodel.set_adrmodel(adrmodel)
        return cmodel
        
    # = fetch Outlier
    def fetch_outlier(self, used_slindexes=None, fitkey=FITKEY, ell_exclusion_to_zero=0.05):
        """ """
        from astropy.stats import mad_std
        ell    = self.get_fitted_value("ell",        slindexes=used_slindexes, fitkey=fitkey)
        ellerr = self.get_fitted_value("ell.err",        slindexes=used_slindexes, fitkey=fitkey)
        theta  = self.get_fitted_value("theta" ,     slindexes=used_slindexes, fitkey=fitkey)
        thetaerr = self.get_fitted_value("theta.err",        slindexes=used_slindexes, fitkey=fitkey)
        # Excluded because boundaries
        flagout = (np.sqrt((ell/1.)**2+(theta/3.14)**2)<ell_exclusion_to_zero) #+ (ellerr<1e-4) + (thetaerr<1e-4) # 1e-4 means ended in boounaries
        ell[flagout] = np.NaN
        theta[flagout] = np.NaN

        flag = np.asarray(flagout + ( np.abs(ell-np.nanmedian(ell))>mad_std(ell[ell==ell])*4 ) + ( np.abs(theta-np.nanmedian(theta))>mad_std(theta[theta==theta])*4),
                              dtype="bool")

        if np.all(flag):
            print("ALL slices have been considered as outlier... set all of them as non-outlier")
            return ~flag
        return flag

        
    # = Get Fitted Parameters
    def get_ellipse_parameters(self, used_slindexes=None, fitkey=FITKEY, exclusion_to_zero=0.05):
        """ estimate the (achromatic) elliptical parameter 

        Returns
        -------
        [mean_ell, mean_ell.err [nMAD], mean_theta, mean_theta.err [nMAD]], mask_removed (True =removed)
        """
        from astropy.stats import mad_std
        if used_slindexes is None:
            used_slindexes = self.slindexes[~self.fetch_outlier()]
            
        ell    = self.get_fitted_value("ell",        slindexes=used_slindexes, fitkey=fitkey)
        theta  = self.get_fitted_value("theta" ,     slindexes=used_slindexes, fitkey=fitkey)

        
        # Excluded because boundaries
        return  [np.average(ell), mad_std(ell)/np.sqrt(len(ell)-1),
                     np.average(theta), mad_std(theta)]
        
    def get_stddev_ratio(self, used_slindexes=None, fitkey=FITKEY):
        """ """
        from astropy.stats import mad_std
        if used_slindexes is None:
            used_slindexes = self.slindexes[~self.fetch_outlier()]
            
        stddev_ratio    = self.get_fitted_value("stddev_ratio",  slindexes=used_slindexes, fitkey=fitkey)
        return np.nanmean(stddev_ratio), mad_std(stddev_ratio)

    def get_amplitude_ratio(self, used_slindexes=None, fitkey=FITKEY):
        """ """
        from astropy.stats import mad_std
        if used_slindexes is None:
            used_slindexes = self.slindexes[~self.fetch_outlier()]
            

        amplitude_ratio    = self.get_fitted_value("amplitude_ratio",  slindexes=used_slindexes, fitkey=fitkey)
        return np.nanmean(amplitude_ratio), mad_std(amplitude_ratio)

    def get_stddev_parameters(self, used_slindexes=None, fitkey=FITKEY, adjust_errors=True, intrinsic=0):
        """ """
        if used_slindexes is None:
            used_slindexes = self.slindexes[~self.fetch_outlier()]
            
        from scipy.optimize import minimize
        lbdas      = self.get_fitted_value("lbda",        slindexes=used_slindexes, fitkey=fitkey)
        stddev     = self.get_fitted_value("stddev",      slindexes=used_slindexes, fitkey=fitkey)
        stddev_err = self.get_fitted_value("stddev.err",  slindexes=used_slindexes, fitkey=fitkey)
        if intrinsic>0: stddev_err = np.sqrt(stddev_err**2 + intrinsic**2)
            
        def _fmin_(param):
            return np.nansum( np.sqrt((stddev-stddev_chromaticity(lbdas, *param))**2/stddev_err**2))

        
        res  = minimize(_fmin_, [np.nanmedian(stddev), -1/5.], bounds=[[0.5,10], [-1,1]], options={"disp":0})
        chi2_dof = res["fun"] / len(stddev-2)
        if chi2_dof>3 and adjust_errors:
            print("Adjusting error")
            from pysedm.utils.tools import fit_intrinsic
            intrinsic = fit_intrinsic(stddev, stddev_chromaticity(lbdas, *res["x"]), stddev_err, len(stddev-2), intrinsic_guess=None)
            return self.get_stddev_parameters( used_slindexes=used_slindexes, fitkey=fitkey, adjust_errors=False,
                                        intrinsic = intrinsic/1.4)
        
        return res["x"]

    
    # Generic
    def get_fitted_value(self, key, slindexes=None, fitkey=FITKEY):
        """ Once the slices has been fitted and recorded, get their fitvalues parameters. 
        
        Parameters
        ----------
        key: [string]
            which `fitvalues` key do you want? (e.g. xcentroid, ell, ...)

        slindexes: [None or list] -optional-
            for which fitted slices do you want that key?
            If None, all the slices wil be used using the self.slindexes property.

        fitkey: [string] -optional-
            Using which key does the `SlicePSF` object created using `fit_slice()` method has been stored.
            [do not change if you don't know]

        Returns
        -------
        1d-array
        """
        if slindexes is None:
            slindexes = self.slindexes
            
        # - Special case, LBDA
        if key in ["lbda","lbdas", "lbdarange"]:
            v = np.asarray([ self.slices[slindex]["lbdarange"] for slindex in slindexes ])
            if key in ["lbda","lbdas"]:
                return np.mean(v, axis=1)
            return v
        
        return np.asarray([self.slices[slindex][fitkey].fitvalues[key] for slindex in slindexes])
    
    # --------- #
    #  SETTER   # 
    # --------- #
    def set_cube(self, cube):
        """ attach a cube to this attribute. """
        self._properties['cube'] = cube

    def load_adrfitter(self, spaxel_unit=1, base_parangle=0):
        """ load the ADRfitter method using the cube's adr.
        This methods need to have the cube loaded (see set_cube())
        """
        from pyifu import adrfit
        if self.cube.adr is None: self.cube.load_adr()
        self._derived_properties['adrfitter'] = adrfit.ADRFitter(self.cube.adr.copy(),
                                                        base_parangle=base_parangle, unit=spaxel_unit)

    def extract_slice(self, slindex, lbda_min=None, lbda_max=None, lbdaindex=None,
                          overwrite=False):
        """ extract a slice from the cube and attach it to the current instance 
        using the index 'slindex'

        = Here level method of `add_slice` = 

        Parameters
        ----------
        slindex: [string/float] 
            Name of the slice. You will recover the extracted slice as self.slices[`slindex`]

        // Slice definition

        lbda_min, lbda_max: [float/None] -optional-
            lower and upper wavelength boundaries [in Angstrom] defining the slice.
            [one of these or lbdaindex must be given]
            
        lbdaindex: [int/None] -optional-
            If you want the slice to be a single wavelength, provide it's index.
            [if this is given, lbda_min, lbda_max are ignored]

        // other
        
        overwrite: [bool] -optional-
            If the slice `slindex` already exists, should this extraction overwrite it?
            
        Returns
        -------
        Void
        """
        if lbda_min is None and lbda_max is None and lbdaindex is None:
            raise ValueError("You need to provide at least one of lbda_min, lbda_max or lbdaindex")

        if lbdaindex is not None:
            lbdarange= [self.cube.lbda[lbdaindex],self.cube.lbda[lbdaindex]]
        else:
            lbdarange = [lbda_min if lbda_min is not None else self.cube.lbda[0],
                         lbda_max if lbda_max is not None else self.cube.lbda[-1]]
                
        # The Slice
        slice_ = self.cube.get_slice(lbda_min=lbda_min, lbda_max=lbda_max,
                                         index=lbdaindex, usemean=True, data='data',
                                         slice_object=True)
        if np.isnan(np.sum(slice_.data)):
            print("psfcube.fitter.py EXTRACT_SLICE: NaN ")
        # - add it
        self.add_slice(slice_, slindex, lbdarange, overwrite=overwrite)

    def add_slice(self, slice_, slindex, lbdarange=None, overwrite=False):
        """ add a new slice to this instance.
    
        The added slice will be accessible as follows:
        ```python
        self.slices[`slindex`] = { 'slice': `slice_`, 'lbdarange': `lbdarange` }
        ```

        Parameters
        ----------
        slice_: [pyifu's Slice]
            The Slice you want to add
            
        slindex: [string/float]
            Name of the slice. You will recover the given  slice as self.slices[`slindex`]

        lbdarange: [float/float] -optional-
            the wavelength range for the slice. This is not mandatory but you should.
            
        // other
        
        overwrite: [bool] -optional-
            If the slice `slindex` already exists, should this extraction overwrite it?
        
        Returns
        -------
        Void
        """
        if slindex in self.slices and not overwrite:
            raise ValueError("slice %d already exists"%slindex)
        
        self.slices[slindex] = {'slice':slice_, 'lbdarange':lbdarange}

    # --------- #
    #  FITTER   # 
    # --------- #    
    # - PSF Slice fitter
    def fit_slice(self, slindex, psfmodel="NormalMoffatTilted",
                    centroids=None, centroids_err=[2,2],
                    adjust_errors=True,
                    fitkey=FITKEY, **kwargs):
        """ fit a PSF on a slice using the fit_slice() function

        Parameters
        ----------
        slindex: [string/float]
            Name of the slice you want to fit.
        
        // PSF Fitting
        
        psfmodel: [string] -optional-
            Name of the model used to fit the PSF e.g.:
            - BiNormalFlat:    PSF + Constant      (1 background param)
            - BiNormalTilted:  PSF + Tilted plane  (3 background params)
            - BiNormalCurved:  PSF + Curved Plane  (5 background params)
        

        centroids: [None/[float,float]] -optional-
            To help the fit, would you have an idea of the PSF centroid?

        centroids_err: [float, float] -optional-
            What would be the -/+ error on you centroid position guess.
            = This is ignored if `centroids` is not provided.

        adjust_errors: [bool] -optional-
            Once the first fit has ran and if the chi2/dof is too high (>2), 
            shall this add an intrinsic dispersion to all points to get closer
            to a chi2/dof and then rerun the fit?
            = you should = 
            
        // other

        fitkey: [string/None] -optional-
            The returned SlicePSF object will be strored as `self.slices[`slindex`][`fitkey`] 
            except if fitkey is None.

        Returns
        -------
        SlicePSF [the object containing the psf fitting methods and results]
        """
        self._test_index_(slindex)
        slpsf = fit_slice(self.slices[slindex]['slice'], psfmodel=psfmodel,
                        centroids=centroids, centroids_err=centroids_err,
                        adjust_errors=adjust_errors, **kwargs)
        
        # - shall this be recorded
        if fitkey is not None:
            self.slices[slindex][fitkey] = slpsf
        
        return slpsf

    # - ADR fitter
    def fit_adr(self, used_slindexes=None, fitkey=FITKEY,
                    parangle=None, spaxel_unit=None,
                    show=False, show_prop={}, 
                     **kwargs):
        """ Fits the adr parameters 

        [This method needs that you have fitted the slices using fit_slice() 
        and stored the results using `fitkey`]
        
        = This method uses pyifu.adrfit (see the load_adrfitter() method ) = 

        Parameters
        ----------
        used_slindexes: [None/list] -optional-
            list of slindex you want to use.
            If None, all the known slices will be used (`self.slindexes`)
            
        fitkey: [string/None] -optional-
            The returned SlicePSF object will be strored as `self.slices[`slindex`][`fitkey`] 
            except if fitkey is None.

        // fit
        spaxel_unit: [float] -optional-
            Size of the spaxels in arcsec. 
            (this parameter is not fitted as this is degenerated with the airmass)
            If not provided during the load_adrfitter, it is suggested that you set it here.

        parangle: [float] -optional-
            Initial guess for the paralactic angle added to the header's one.
            Note: **kwargs goes to `adrfitter.fit()` as modefit fit properties.
        
        // other

        show: [bool] -optional-
            Shall this plot the results.
        
        show_prop: [dict] -optional-
            dictionary sent as kwargs for adrfitter.show(**show_prop)
            e.g.: {"ax":ax, "show":False}

        
        **kwargs goes to adrfitter.fit() [modefit fit prop kwargs]

        Returns
        -------
        dict (fitvalues)
        """

        if used_slindexes is None:
            used_slindexes = self.slindexes[~self.fetch_outlier()]

        lbda  = np.mean([self.slices[slindex]['lbdarange'] for slindex in used_slindexes], axis=1)
        x0    = self.get_fitted_value("xcentroid",slindexes=used_slindexes,     fitkey=fitkey)
        x0err = self.get_fitted_value("xcentroid.err",slindexes=used_slindexes, fitkey=fitkey)
        y0    = self.get_fitted_value("ycentroid",slindexes=used_slindexes,     fitkey=fitkey)
        y0err = self.get_fitted_value("ycentroid.err",slindexes=used_slindexes, fitkey=fitkey)
        if spaxel_unit is not None: self.adrfitter.model._unit = spaxel_unit
        self.adrfitter.set_data(lbda, x0, y0, x0err, y0err)
        
        
        if parangle is None:
            parangle_guess = self.cube.header["TEL_PA"]+10 # + 10 because of exposure time drifting
        else:
            parangle_guess = parangle
            
        default_guesses = dict(airmass_guess=self.cube.header["AIRMASS"]+0.05, # for drifting
                               airmass_boundaries=[1.0005,self.cube.header["AIRMASS"]*1.4],
                               xref_guess= np.mean(x0), yref_guess= np.mean(y0),
                               parangle_guess=parangle_guess,
                               parangle_boundaries=[parangle_guess-270,parangle_guess+270])

        self.adrfitter.fit( **kwargs_update(default_guesses,**kwargs) )
        
        if self.adrfitter.dof> 0 and self.adrfitter.fitvalues["chi2"] / self.adrfitter.dof >5:
            print("WARNING: ADR fit chi2/dof of %.1f - most likely a badly fitted point is causing trouble"%(self.adrfitter.fitvalues["chi2"] / self.adrfitter.dof))
            
        if show:
            self.adrfitter.show(**show_prop)
            
        return self.adrfitter.fitvalues


    # --------- #
    # PLOTTING  # 
    # --------- #
    def show_ellipse(self, used_slindexes=None, show_model=True):
        """ """
        
        if used_slindexes is None:
            used_slindexes = self.slindexes
            
        mask_removed = self.fetch_outlier(used_slindexes=used_slindexes)
        kept_slindexes = np.asarray(used_slindexes)[~mask_removed]
        rejected_slindexes = np.asarray(used_slindexes)[mask_removed]
        
        #                 #
        #    Data         #
        #                 #
        [mean_ell, mean_ellerr, mean_theta, mean_thetaerr]  = self.get_ellipse_parameters(used_slindexes=kept_slindexes)
        #                 #
        #    Axis         #
        #                 #
        fig  = mpl.figure(figsize=[6,6])
        
        error_prop   = dict(ls="None", marker="None", ms=0, ecolor="0.7", zorder=1)
        scatter_prop = dict(s=50, zorder=4)

        
        
        fig = self._show_corner_( ["ell","theta"], fig=fig, labels=["ellipticity",r"Angle [rad]"],
                                  expectation=[mean_ell, mean_theta] if show_model else None, expectation_err=[mean_ellerr, mean_thetaerr],
                                  used_slindexes = kept_slindexes,
                                  error_prop=error_prop, **scatter_prop )
        
        if np.any(mask_removed):
            scatter_prop_out = dict(s=20, zorder=4, facecolors="None", edgecolors="0.7")
            fig = self._show_corner_( ["ell","theta"], fig=fig, labels=["ellipticity",r"Angle [rad]"],
                                  used_slindexes = rejected_slindexes, show_labels=False,
                                  error_prop=error_prop, **scatter_prop_out )
        
    def show_profile(self, used_slindexes=None, show_model=True,
                         psfmodel="NormalMoffat"):
        """ """
        if used_slindexes is None:
            used_slindexes = self.slindexes
            
        mask_removed = self.fetch_outlier(used_slindexes=used_slindexes)
        kept_slindexes = np.asarray(used_slindexes)[~mask_removed]
        rejected_slindexes = np.asarray(used_slindexes)[mask_removed]

        if psfmodel in ["BiNormal"]:
            profile_parameters = ["stddev","stddev_ratio","amplitude_ratio"]
            labels = ["Scale [std]",r"Scale ratio","Amplitude Ratio"]
        elif psfmodel in ["NormalMoffat", "MoffatNormal"]:
            profile_parameters = ["stddev","alpha","amplitude_ratio"]
            labels = ["Scale [std]",r"Moffat $\alpha$","Amplitude Ratio"]
            show_model = False
        elif psfmodel in ["Moffat"]:
            profile_parameters = ["alpha","beta"]
            labels = [r"Moffat $\alpha$",r"Moffat $\beta$"]
            show_model = False
        else:
            raise ValueError("only BiNormal and NormalMoffat implemented")
        #                 #
        #    Data         #
        #                 #
        if show_model:
            amplitude_ratio, amplitude_ratioerr = self.get_amplitude_ratio( used_slindexes=kept_slindexes)
            stddev_ratio, stddev_ratioerr       = self.get_stddev_ratio(    used_slindexes=kept_slindexes)
        #                 #
        #    Axis         #
        #                 #
        fig  = mpl.figure(figsize=[6,6])
        error_prop   = dict(ls="None", marker="None", ms=0, ecolor="0.7", zorder=1)
        scatter_prop = dict(s=50, zorder=4)
        expectation_color = "C5"
        #   The Figure    #
        fig = self._show_corner_( profile_parameters, fig=fig,
                                  labels=labels,
                                  used_slindexes = kept_slindexes,
                                  expectation=[None,stddev_ratio, amplitude_ratio] if show_model else None,
                                  expectation_err=[None,stddev_ratioerr, amplitude_ratioerr] if show_model else None,
                                  expectation_color=expectation_color,
                                  error_prop=error_prop, **scatter_prop )


        if np.any(mask_removed):
            scatter_prop_out = dict(s=20, zorder=4, facecolors="None", edgecolors="0.7")
            fig = self._show_corner_(  profile_parameters, fig=fig,
                                           used_slindexes = rejected_slindexes,
                                       labels=[None]*len(profile_parameters), show_labels=False,
                                        error_prop=error_prop, **scatter_prop_out )

            
        if show_model:
            lbda = np.linspace(3000,10000,100)
            ax1 = fig.axes[0]
            stddevref, rho = self.get_stddev_parameters(used_slindexes=used_slindexes)
            ax1.plot(lbda, stddev_chromaticity(lbda, stddevref, rho=rho), color=expectation_color,
                         scalex=False)
            ax1.plot(lbda, stddev_chromaticity(lbda, stddevref, rho=-1/5), color="0.5", ls="--", alpha=0.5, zorder=1,
                         scalex=False)
            ax1.text(0.9,0.9, "rho=%.2f"%rho, color=expectation_color,
                         va="top",ha="right", transform = ax1.transAxes)

    def show_adr(self, ax=None, **kwargs):
        """ """
        self.adrfitter.show(ax=ax, guess_airmass=self.cube.header["AIRMASS"], **kwargs)
        
    # =================== #
    #   Internal          #
    # =================== #
    def _show_kx_v_ky_(self, ax, keyx, keyy, keyxerr=None, keyyerr=None, used_slindexes=None,
                           error_prop={}, **kwargs):
        """ """
        x_ = self.get_fitted_value(keyx, slindexes=used_slindexes)
        dx_ = None if keyxerr is None else self.get_fitted_value(keyxerr, slindexes=used_slindexes)
        y_ = self.get_fitted_value(keyy, slindexes=used_slindexes)
        dy_ = None if keyyerr is None else self.get_fitted_value(keyyerr, slindexes=used_slindexes)
        
        ax.scatter(x_, y_, **kwargs_update( dict(s=50, zorder=4),**kwargs))
        ax.errorbar(x_, y_, xerr=dx_, yerr=dy_,
                        **kwargs_update( dict(ls="None", marker="None", ms=0, ecolor="0.7", zorder=1),**error_prop))
        
    def _show_corner_(self,  parameters, fig=None, used_slindexes=None, error_prop={},
                          expectation=None, expectation_err=None, expectation_color="C5",
                          show_labels=True,
                          labels=None,**kwargs):
        """ """
        n_param = len(parameters)
        if fig is None:
            fig = mpl.figure()

        if labels is None:
            labels = parameters
        if expectation is None: expectation = [None] * n_param
                
        if expectation_err is None: expectation_err = [None] * n_param
                
        for i,xkey in enumerate(parameters):
            for j,ykey in enumerate(parameters):
                if j>i : continue
                ax = fig.add_subplot(n_param,n_param,i*n_param+(j+1))
                if j ==0 and show_labels: ax.set_ylabel(labels[i])
                    
                if ykey == xkey:
                    self._show_kx_v_ky_(ax, "lbda",ykey, keyyerr=ykey+".err",
                                used_slindexes=used_slindexes, error_prop=error_prop, **kwargs)
                    if show_labels: ax.set_xlabel(r"Wavelength [$\AA$]")
                    if expectation[i] is not None:
                        ax.axhline(expectation[i], color=expectation_color)
                        if expectation_err[i] is not None:
                            ax.axhspan(expectation[i]-expectation_err[i],expectation[i]+expectation_err[i],
                                    color=expectation_color, alpha=0.3)
                else:
                    self._show_kx_v_ky_(ax,ykey,xkey, keyxerr=ykey+".err", keyyerr=xkey+".err",
                                    used_slindexes=used_slindexes, error_prop=error_prop, **kwargs)
                    if expectation[i] is not None and expectation[j] is not None:
                        ax.scatter(expectation[j], expectation[i], color=expectation_color,
                                       **kwargs_update(kwargs,**{"marker":"s"}))
                        if expectation_err[i] is not None or expectation_err[j] is not None:
                            ax.errorbar(expectation[j], expectation[i],
                                        xerr=expectation_err[j], yerr=expectation_err[i], 
                                        **kwargs_update(error_prop, **{"ecolor":expectation_color}))
                            
                    if i==n_param-1 and show_labels: ax.set_xlabel(labels[j])
        fig.tight_layout()
        return fig
    # =================== #
    #   Properties        #
    # =================== #
    def _test_index_(self, slindex):
        """ Raises a ValueError if index not in self.slices """
        if slindex not in self.slices:
            raise ValueError("unknown slice '%s'"%slindex)

    # -------
    #  Slice
    # ------- 
    @property
    def slices(self):
        """ """
        if self._properties['slices'] is None:
            self._properties['slices'] = {}
        return self._properties['slices']


    @property
    def slindexes(self):
        """ """
        return np.asarray(list(self.slices.keys()))
    
    # -------
    #  Cube
    # ------- 
    @property
    def cube(self):
        """ """
        return self._properties['cube']

    # ========= #
    # -------
    # ADR
    # -------
    @property
    def adrfitter(self):
        """ """
        return self._derived_properties["adrfitter"]


################################
#                              #
#                              #
#     SLICE FITTER             #
#                              #
#                              #
################################
class PSFFitter( BaseFitter ):
    """ """
    Properties         = ["spaxelhandler"]
    SIDE_PROPERTIES    = ["fit_area","errorscale","intrinsicerror"]
    DERIVED_PROPERTIES = ["fitted_indexes","dataindex",
                          "xfitted","yfitted","datafitted","errorfitted"]
    # -------------- #
    #  SETTER        #
    # -------------- #
    def _set_spaxelhandler_(self, spaxelhandler ) :
        """ """
        self._properties["spaxelhandler"] = spaxelhandler
        
    def set_fit_area(self, polygon):
        """ Provide a polygon. Only data within this polygon will be fit 

        Parameters
        ----------
        polygon: [shapely.geometry.Polygon or array]
            The polygon definition. Spaxels within this area will be fitted.
            This could have 2 formats:
            - array: the vertices. The code will create the polygon using shapely.geometry(polygon)
            - Polygon: i.e. the result of shapely.geometry(polygon)
        
        Returns
        -------
        Void
        """
        if type(polygon) in [np.array, np.ndarray, list]:
            polygon = shapely.geometry(polygon)
        
        self._side_properties['fit_area'] = polygon
        self.set_fitted_indexes(self._spaxelhandler.get_spaxels_within_polygon(polygon))
        
    def set_fitted_indexes(self, indexes):
        """ provide the spaxel indexes that will be fitted """
        self._derived_properties["fitted_indexes"] = indexes
        self._set_fitted_values_()
       
    # ================ #
    #  Properties      #
    # ================ #
    @property
    def _spaxelhandler(self):
        """ """
        return self._properties['spaxelhandler']


    def _set_fitted_values_(self):
        """ """
        x, y = np.asarray(self._spaxelhandler.index_to_xy(self.fitted_indexes)).T
        self._derived_properties['xfitted'] = x
        self._derived_properties['yfitted'] = y
        self._derived_properties['datafitted']  = self._spaxelhandler.data.T[self._fit_dataindex].T
        if USE_LEASTSQ:
            from astropy.stats import mad_std
            
            self._derived_properties['errorfitted'] = mad_std(self._datafitted[self._datafitted==self._datafitted])*1.4
            self.set_error_scale(1)
            self.set_intrinsic_error(0)
            
        else:
            if np.any(self._spaxelhandler.variance.T[self._fit_dataindex]<0):
                warnings.warn("Negative variance detected. These variance at set back to twice the median vairance.")
                var = self._spaxelhandler.variance.T[self._fit_dataindex]
                var[var<=0] = np.nanmedian(var)*2
                self._derived_properties['errorfitted'] = np.sqrt(var)
            else:
                self._derived_properties['errorfitted'] = np.sqrt(self._spaxelhandler.variance.T[self._fit_dataindex]).T
            
            if self._side_properties['errorscale'] is None:
                self.set_error_scale(1)
            
            if self._side_properties['intrinsicerror'] is None:
                self.set_intrinsic_error(0)

    def set_error_scale(self, scaleup):
        """ """
        self._side_properties['errorscale']  = scaleup

    def set_intrinsic_error(self, int_error):
        """ """
        self._side_properties['intrinsicerror'] = int_error
        
    @property
    def _intrinsic_error(self):
        """ """
        return self._side_properties['intrinsicerror']
        
    @property
    def _xfitted(self):
        """ """
        return self._derived_properties['xfitted']
    @property
    def _yfitted(self):
        """ """
        return self._derived_properties['yfitted']
    @property
    def _datafitted(self):
        """ """
        return self._derived_properties['datafitted']
    
    @property
    def _errorfitted(self):
        """ """
        return self._derived_properties['errorfitted'] * self._errorscale + self._intrinsic_error

    @property
    def _errorscale(self):
        """ """
        return self._side_properties['errorscale']
    
    # - indexes and ids
    @property
    def fit_area(self):
        """ polygon of the restricted fitted area (if any) """
        return self._side_properties['fit_area']

    @property
    def fitted_indexes(self):
        """ list of the fitted indexes """
        if self._derived_properties["fitted_indexes"] is None:
            return self._spaxelhandler.indexes
        return self._derived_properties["fitted_indexes"]
    
    @property
    def _fit_dataindex(self):
        """ indices associated with the indexes """
        
        if self._derived_properties["fitted_indexes"] is None:
            return np.arange(self._spaxelhandler.nspaxels)
        # -- Needed to speed up fit
        if self._derived_properties["dataindex"] is None:
            self._derived_properties["dataindex"] = \
              np.in1d( self._spaxelhandler.indexes, self.fitted_indexes)
              
        return self._derived_properties["dataindex"]

class SlicePSF( PSFFitter ):
    """ """
    # =================== #
    #   Methods           #
    # =================== #
    def __init__(self, slice_,
                     fitbuffer=None,fit_area=None,
                     psfmodel="NormalMoffatTilted",
                     fitted_indexes=None):
        """ The SlicePSF fitter object

        Parameters
        ---------- 
        slice_: [pyifu Slice] 
            The slice object that will be fitted
            

        fitbuffer: [float] -optional- 
            = Ignored if fit_area or fitted_indexes are given=

        psfmodel: [string] -optional-
            Name of the PSF model used to fit the slice. 
            examples: 
            - MoffatPlane`N`:a Moffat2D profile + `N`-degree Polynomial2D background 
        
        """
        self.set_slice(slice_)
        # - Setting the model
        self.set_model(read_psfmodel(psfmodel))

        # = Which Data
        if fitted_indexes is not None:
            self.set_fitted_indexes(fitted_indexes)
        elif fit_area is not None:
            self.set_fit_area(fit_area)
        elif fitbuffer is not None:
            self._set_fitted_values_()
            g = self.get_guesses() 
            x,y = self.model.centroid_guess
            self.set_fit_area(shapely.geometry.Point(x,y).buffer(fitbuffer))
        else:
            self._set_fitted_values_()
            
        self.use_minuit = True

    # --------- #
    #  FITTING  #
    # --------- #
    def _get_model_args_(self):
        """ see model.get_loglikelihood"""
        self._set_fitted_values_()
        # corresponding data entry:
        return self._xfitted, self._yfitted, self._datafitted, self._errorfitted

    def get_guesses(self, xcentroid=None, xcentroid_err=2, ycentroid=None, ycentroid_err=2):
        """ you can help to pick the good positions by giving the x and y centroids """
        return self.model.get_guesses(self._xfitted, self._yfitted, self._datafitted,
                            xcentroid=xcentroid, xcentroid_err=xcentroid_err,
                            ycentroid=ycentroid, ycentroid_err=ycentroid_err)

    # --------- #
    #  SETTER   #
    # --------- #
    def set_slice(self, slice_):
        """ set a pyifu slice """
        #from pyifu.spectroscopy import Slice
        #if Slice not in slice_.__class__.__mro__:
        #   raise TypeError("the given slice is not a pyifu Slice (of Child of)")
        self._set_spaxelhandler_(slice_)
        
    # --------- #
    # PLOTTER   #
    # --------- #
    def show_psf(self, ax=None, show=True, savefile=None, nobkgd=True, **kwargs):
        """ """
        import matplotlib.pyplot as mpl
        from .model import get_elliptical_distance
        if ax is None:
            fig = mpl.figure(figsize=[6,4])
            ax  = fig.add_axes([0.13,0.1,0.77,0.8])
        else:
            fig = ax.figure
            
            
        r_ellipse = get_elliptical_distance(self._xfitted, self._yfitted,
                                                  xcentroid=self.fitvalues['xcentroid'],
                                                  ycentroid=self.fitvalues['ycentroid'],
                                                  ell=self.fitvalues['ell'], theta=self.fitvalues['theta'])
        if nobkgd:
            background = self.model.get_background(self._xfitted, self._yfitted)
            datashown = self._datafitted - background
        else:
            datashown = self._datafitted
        ax.scatter(r_ellipse, datashown, marker="o", zorder=2, s=80, edgecolors="0.7",
                       facecolors=mpl.cm.binary(0.2,0.7))
        ax.errorbar(r_ellipse, datashown, yerr=self._errorfitted,
                    marker="None", ls="None", ecolor="0.7", zorder=1, alpha=0.7)

        
        self.model.display_model(ax, np.linspace(0.0,np.nanmax(r_ellipse),500), nobkgd=nobkgd,
                                     **kwargs)
        
        if savefile:
            fig.savefig(savefile)
        if show:
            fig.show()
        
    def show(self, savefile=None, show=True,
                 centroid_prop={}, logscale=True,psf_in_log=True, 
                 vmin="2", vmax="98", ylim_low=None, xlim=[0,10], **kwargs):
        """ """
        import matplotlib.pyplot            as mpl
        from .tools     import kwargs_update
        from pyifu.spectroscopy import get_slice
        
        # -- Axes Definition
        fig = mpl.figure(figsize=(9, 2.5))
        left, width, space = 0.05, 0.15, 0.02
        bottom, height = 0.2, 0.65
        axdata  = fig.add_axes([left+0*(width+space), bottom, width, height])
        axmodel = fig.add_axes([left+1*(width+space), bottom, width, height],
                                   sharex=axdata, sharey=axdata)
        axres   = fig.add_axes([left+2*(width+space), bottom, width, height],
                                   sharex=axdata, sharey=axdata)
        
        axpsf   = fig.add_axes([left+3*(width+space)+space*1.5, bottom, 0.955-(left+3*(width+space)+space), height])

        # = Data
        slice_    = self.slice.data 
        slice_var = self.slice.variance 
        x,y       = np.asarray(self.slice.index_to_xy(self.slice.indexes)).T
        model_    = self.model.get_model(x ,y)
        model_slice = get_slice(model_, np.asarray(self.slice.index_to_xy(self.slice.indexes)),
                                    spaxel_vertices=self.slice.spaxel_vertices, variance=None,
                                    indexes=self.slice.indexes)
        
        res_      = slice_ - model_
        res_slice = get_slice(res_, np.asarray(self.slice.index_to_xy(self.slice.indexes)),
                                    spaxel_vertices=self.slice.spaxel_vertices, variance=None,
                                    indexes=self.slice.indexes)
        # = Plot
        self.slice.show( ax=axdata, vmin=vmin, vmax=vmax , show_colorbar=False, show=False)
        model_slice.show( ax=axmodel, vmin=vmin, vmax=vmax , show_colorbar=False, show=False )
        res_slice.show( ax=axres, vmin=vmin, vmax=vmax , show_colorbar=False, show=False )
        self.show_psf(ax=axpsf, show=False, scalex=False, scaley=False)
        
        # fancy
        [ax_.set_yticklabels([]) for ax_ in fig.axes[1:]]
        axdata.set_title("Data")
        axmodel.set_title("Model")
        axres.set_title("Residual")
            
        axpsf.set_xlabel("Elliptical distance [in spaxels]")
        if xlim is not None:
            axpsf.set_xlim(*xlim)
            
        if psf_in_log:
            if ylim_low is None: ylim_low = 1
            axpsf.set_ylim(ylim_low, slice_.max()*2)
            axpsf.set_yscale("log")

        fig.text(0.95,0.95, "model: %s"%self.model.NAME, fontsize="small",
                     va="top", ha="right")
        fig.figout(savefile=savefile, show=show)
        
    # =================== #
    #  Properties         #
    # =================== #
    @property
    def slice(self):
        """ pyifu slice """
        return self._spaxelhandler

    @property
    def npoints(self):
        """ """
        return len(self._datafitted)

    @property
    def lbda(self):
        """ wavelength of the fitted slice (if given) """
        return self.slice.lbda



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""  """


import numpy as np



def extract_star(cube, lbda_step1=None, psfmodel="NormalMoffatTilted",
                centroids=None, centroids_err=[5,5],
                only_step1=False, spaxel_unit=1, step1_fit_prop={},
                final_slice_width=None,
                force_ellipse=True, force_centroid=True, force_stddev=True, force_alpha=True,
                normalized=False, ncore=None, notebook=False, verbose=True):
    """ 
    Returns
    -------
    spectrum, model (cube), psfmodel (cube), bkgdmodel (cube)
    """
    
    from pyifu   import get_spectrum, get_cube
    
    if lbda_step1 is None:
        lbdaranges, bins = [5000,8000], 6
        STEP_LBDA_RANGE = np.linspace(lbdaranges[0],lbdaranges[1], bins+1)
        lbda_step1      = np.asarray([STEP_LBDA_RANGE[:-1], STEP_LBDA_RANGE[1:]]).T

        
    # Step 1

    psffit = fit_metaslices(cube, lbda_step1, 
                            psfmodel=psfmodel, 
                            centroids=centroids, centroids_err=centroids_err,
                            spaxel_unit=spaxel_unit, **step1_fit_prop)
    if only_step1:
        return psffit
    
    # Step 2

    # ellipse_parameters
    ell, ellerr, theta, thetaerr = psffit.get_ellipse_parameters()
    
    cmodel = psffit.get_chromatic_profile_model()
    
    slfits = cmodel.force_fit(cube, ell=ell, theta=theta,
                                  ellerr=ellerr*2, thetaerr=thetaerr*2,
                                  psfmodel=psfmodel,
                                  force_ellipse=force_ellipse,
                                  force_centroid=force_centroid,
                                  force_stddev=force_stddev, force_alpha=force_alpha,
                                  slice_width=final_slice_width,
                                  )
    lbdas = np.asarray([slfits[i].lbda for i in range(len(slfits))])
    # Returns all structures
    cube_prop = dict(header=cube.header, lbda=lbdas,
                    spaxel_mapping = cube.spaxel_mapping, spaxel_vertices=cube.spaxel_vertices)

    # Background
    databkgd = np.asarray([slfits[i].model.get_background(slfits[i]._xfitted, slfits[i]._yfitted)
                                           for i in range(len(lbdas))])
    if len(np.shape(databkgd)) ==1: # means "Flat"
        databkgd = np.asarray([databkgd for i in range(len(cube.indexes))]).T
        
    bkgdmodel = get_cube(  databkgd, **cube_prop)
    # PSF
    psfmodel_  = get_cube(  np.asarray([slfits[i].model.get_profile(slfits[i]._xfitted, slfits[i]._yfitted)
                                           for i in range(len(lbdas))]),
                        **cube_prop)
    # Complit Model
    model     = get_cube(  np.asarray([slfits[i].model.get_model(slfits[i]._xfitted, slfits[i]._yfitted)
                                           for i in range(len(lbdas))]),
                        **cube_prop)
    # = The spectrum
    flux,err  = np.asarray([[slfits[i].fitvalues["amplitude"],slfits[i].fitvalues["amplitude.err"]]  for i in range(len(lbdas))]).T
    
    # Normalization
    if normalized:
        print(" ******************* ")
        print("WARNING: normalized tools is not accurate yet.")
        print(" ******************* ")
        normalization = _get_spectrum_normalization_(slfits, psfmodel=psfmodel,
                                                         ncore=ncore, notebook=notebook, verbose=verbose)
        norm = np.asarray([normalization[i][0] for i in range( len(normalization) ) ])
        flux /= norm
        err  /= norm
        
    spectrum  = get_spectrum(lbdas, flux, variance=err**2, header=cube.header)
    return spectrum, model, psfmodel_, bkgdmodel, psffit, slfits


def _get_spectrum_normalization_(slfits, ncore=None, notebook=False, verbose=True,
                                     psfmodel="NormalMoffatTilted"):
    """ """
    if "NormalMoffat" in psfmodel:
        from .model import get_normalmoffat_normalisation
        normalisation_func = get_normalmoffat_normalisation
    else:
        raise NotImplementedError("Only NormalMoffat profile normalization implemented")
    
    import multiprocessing
    from astropy.utils.console import ProgressBar
    if ncore is None:
        if multiprocessing.cpu_count()>20:
            ncore = multiprocessing.cpu_count() - 8
        elif multiprocessing.cpu_count()>8:
            ncore = multiprocessing.cpu_count() - 5
        else:
            ncore = multiprocessing.cpu_count() - 2
        if ncore==0:
            ncore = 1

    if verbose: print("Measuring Spectrum Normalization, using %d cores"%ncore)
    with ProgressBar( len(slfits), ipython_widget=notebook) as bar:
        p = multiprocessing.Pool(ncore)
        res = {}
        for j, result in enumerate( p.imap( normalisation_func, [slfits[i].model.param_profile for i in range(len( slfits))] )):
            res[j] = result
            bar.update(j)
                
    return res
        
        


    
def build_parameter_prior(filenames, centroids=None, psfmodel="NormalMoffatTilted", lbdaranges=[4000,8500], bins=10):
    """ """
    prop_fit = {}

    STEP_LBDA_RANGE = np.linspace(lbdaranges[0],lbdaranges[1], bins+1)
    lbdas           = np.asarray([STEP_LBDA_RANGE[:-1], STEP_LBDA_RANGE[1:]]).T
    
    def _fit_cube_(cube):
        psffit = SlicePSFCollection()
        psffit.set_cube(cube)
        for i,lbdar in enumerate(lbdas):
            psffit.extract_slice(i, *lbdar)
            slpsf = psffit.fit_slice(i, psfmodel=psfmodel,
                                        **prop_fit)
            return slpsf
        
    return {filename_: _fit_cube_(pysedm.get_sedmcube(filename_)) for filename_ in filenames}


def fit_metaslices(cube, lbdas, psfmodel="NormalMoffatTilted",
                       centroids=None, centroids_err=[5,5],
                       spaxel_unit=1,
                       **kwargs):
    """ """
    from .fitter import SlicePSFCollection
    psffit = SlicePSFCollection()
    psffit.set_cube(cube)

    for i,lbdar in enumerate(lbdas):
        psffit.extract_slice(i, *lbdar)
        slpsf = psffit.fit_slice(i, psfmodel=psfmodel,
                                       centroids=centroids,
                            centroids_err=centroids_err, **kwargs)

    psffit.load_adrfitter(spaxel_unit=spaxel_unit)
    psffit.fit_adr()
    return psffit

def automatic_fit_psf(cube, centroids=[0,0],
                        centroids_err=[3,3],
                        psfmodel="NormalMoffatTilted", step_bins=[3,10]):
    """ """
    prop_fit = dict(ell_boundaries=[0.01,0.5])
    # ================== #
    # Step 1: 5 slices  #
    #  All Free          #
    # ================== #
    STEP1_LBDA_RANGE = np.linspace(4500,8000, step_bins[0]+1)
    step1_lbdas = np.asarray([STEP1_LBDA_RANGE[:-1], STEP1_LBDA_RANGE[1:]]).T

    psffit_step1 = SlicePSFCollection()
    psffit_step1.set_cube(cube)

    for i,lbdar in enumerate(step1_lbdas):
        psffit_step1.extract_slice(i, *lbdar)
        slpsf = psffit_step1.fit_slice(i, psfmodel=psfmodel,
                                       centroids=centroids,
                            centroids_err=centroids_err, **prop_fit)

    psffit_step1.load_adrfitter()
    psffit_step1.fit_adr()
    if len(step_bins)==1 or step_bins[1] is None:
        return psffit_step1, None
    
    # ================== #
    # Step 2: 15 slices  #
    #  Strong centroid   #
    # ================== #
    # - Helping on the ellipticity
    [mean_ell, mean_ellerr, mean_theta, mean_thetaerr], mask_removed  = psffit_step1.get_ellipse_parameters()
    stddev_ratio,stddev_ratioerr = psffit_step1.get_stddev_ratio()
    
    STEP2_LBDA_RANGE = np.linspace(4500,8000, step_bins[1]+1)
    step2_lbdas = np.asarray([STEP2_LBDA_RANGE[:-1], STEP2_LBDA_RANGE[1:]]).T
    STEP2_CENTROID_ERROR = [0.2, 0.2]
    
    psffit_step2 = SlicePSFCollection()
    psffit_step2.set_cube(cube)

    prop_fit["ell_guess"] = mean_ell
    prop_fit["ell_boundaries"] = [mean_ell-mean_ellerr, mean_ell+mean_ellerr]

    prop_fit["theta_guess"] = mean_theta
    prop_fit["theta_boundaries"] = [mean_theta-mean_thetaerr, mean_theta+mean_thetaerr]

    prop_fit["stddev_ratio_guess"] = stddev_ratio
    prop_fit["stddev_ratio_boundaries"] = [stddev_ratio-stddev_ratioerr, stddev_ratio+stddev_ratioerr]
    
    
    for i,lbdar in enumerate(step2_lbdas):
        centroids = psffit_step1.get_adr_centroid( np.mean(lbdar) )
        psffit_step2.extract_slice(i, *lbdar)
        slpsf = psffit_step2.fit_slice(i, psfmodel=psfmodel,
                                        centroids=centroids,
                                        centroids_err=STEP2_CENTROID_ERROR, **prop_fit)

    psffit_step2.load_adrfitter()
    psffit_step2.fit_adr()

    # ================== #
    # Step 2: 15 slices  #
    #  Strong centroid   #
    # ================== #


    return psffit_step1, psffit_step2



#! /usr/bin/env python
# -*- coding: utf-8 -*-

""" PSF Modelisation """


import numpy as np
from scipy.stats import norm
from scipy import integrate
from modefit.baseobjects import BaseModel

################################
#                              #
#                              #
#  Main User Functions         #
#                              #
#                              #
################################


def read_psfmodel(psfmodel):
    """ """
    
    if "BiNormalFlat" in psfmodel:
        return BiNormalFlat()
    elif "BiNormalTilted" in psfmodel:
        return BiNormalTilted()
    elif "BiNormalCurved" in psfmodel:
        return BiNormalCurved()
    elif "NormalMoffatFlat" in psfmodel:
        return NormalMoffatFlat()
    elif "NormalMoffatTilted" in psfmodel:
        return NormalMoffatTilted()
    elif "NormalMoffatCurved" in psfmodel:
        return NormalMoffatCurved()
    
    elif "MoffatFlat" in psfmodel:
        return MoffatFlat()
    elif "MoffatTilted" in psfmodel:
        return MoffatTilted()
    elif "MoffatCurved" in psfmodel:
        return MoffatCurved()

    else:
        raise ValueError("Only the '{BiNormal/NormalMoffat}{Flat/Tilted/Curved}' psfmodel has been implemented")


################################
#                              #
#                              #
#  Profile and Background      #
#                              #
#                              #
################################

# ========================= #
#                           #
#  Profiles                 #
#                           #
# ========================= #
"""
# How To build a new profile?

All profile must have the following format:
```python
def name_of_the_profile(x,y,
                        here, define, your, variables,
                        xcentroid=0, ycentroid=0):
     ''' ADD Documentation '''

     do_the_job

     return list_of_values (1darray with the same length as x/y)
```
The profile should be normalize, but this is not mendatory.

"""
def binormal_profile(x, y,
                stddev, stddev_ratio, amplitude_ratio, theta, ell,
                xcentroid=0, ycentroid=0,
                amplitude=1):
    """ Return the model profile.
    Here a binormal profile.
    if decomposed
        
    Returns
    -------
    array (model)
            
    """
    
    r = get_elliptical_distance(x, y, xcentroid=xcentroid, ycentroid=ycentroid,  ell=ell, theta=theta)
    
    n1 = _normal_(r, scale=stddev)
    n2 = _normal_(r, scale=stddev*stddev_ratio)

    coef1 = amplitude_ratio/(1.+amplitude_ratio)
    coef2 = 1./(1+amplitude_ratio)
    
    return amplitude * ( coef1 * n1 + coef2 * n2)

def moffat_profile(x, y, alpha, beta,
                       theta, ell,
                       xcentroid=0, ycentroid=0,
                       amplitude=1):
    """ Return the model profile.
    Here a moffat profile.

    if decomposed
        
    Returns
    -------
    array (model)
    """
    r = get_elliptical_distance(x, y, xcentroid=xcentroid, ycentroid=ycentroid,  ell=ell, theta=theta)
    n1 = _moffat_(r, alpha, beta)
    return amplitude * ( n1 )


def normalmoffat_profile(x, y,
                        stddev, alpha, amplitude_ratio, theta, ell,
                        xcentroid=0, ycentroid=0,
                        amplitude=1):
    """ Return the model profile.
    Here a normal + moffat profile.
    if decomposed
        
    Returns
    -------
    array (model)
    """
    
    r = get_elliptical_distance(x, y, xcentroid=xcentroid, ycentroid=ycentroid,  ell=ell, theta=theta)
    n1 = _normal_(r, scale=stddev) 
    n2 = _moffat_(r, alpha, _alpha_to_beta_(alpha)) 

    coef1 = amplitude_ratio/(1.+amplitude_ratio)
    coef2 = 1./(1+amplitude_ratio)
    
    return amplitude * ( coef1 * n1 + coef2 * n2 )

def _alpha_to_stddev_(alpha, sigma0=-0.1, sigma1=1.4 ):
    """ """
    return sigma0+alpha*sigma1

def _alpha_to_beta_(alpha, b0=0.25, b1=0.63):
    """ Ratio given by SNIFS """
    return b0+alpha*b1

# Profiles 
def _normal_(r, scale):
    """ """
    return norm.pdf(r, loc=0, scale=scale)

def _moffat_(r, alpha, beta):
    """ """
    return  1/(2*_default_moffat_normalization_(alpha)) * (1 + (r/alpha)**2 )**(-beta)

# profile_normalization
def _default_moffat_normalization_(alpha):
    """ To be used when using _alpha_to_beta_ """
    coefs = [1.6532341084340385,
                 0.0572839305590741,
                 0.08240094225731157,
                 -0.01578126134847564,
                 0.0013096135691818885,
                 -4.1053249828530274e-05]
    return np.dot([1,alpha,alpha**2, alpha**3, alpha**4, alpha**5],coefs)



def get_normalmoffat_normalisation( param_profile,
                                    xbounds=50, ybounds=50,
                                    epsabs=1e-2 ):
    """ measures the normalisation coefficient one should apply to fitvalues["amplitude"] to have 
    the effective amplitude (i.e., integral of the 2D profile model is 1 if "amplitude" equals 1).
    
    
    Parameters
    ----------

    epsabs : float, optional
        Absolute tolerance passed directly to the inner 1-D quadrature integration. 
        # Scipy default is 1.49e-8.



    // from scipy.integrate.dblquad  //

    a, b : float
        The limits of integration in x: a < b

        => a =  xcentroid - xbound
        => b =  xcentroid + xbound

    gfun : callable or float
        The lower boundary curve in y which is a function taking a single floating point argument (x) and returning a floating point result or a float indicating a constant boundary curve.

        => gfun = lambda x:  ycentroid-ybounds

    hfun : callable or float
        The upper boundary curve in y (same requirements as gfun).

        => hfun = lambda x:  ycentroid+ybounds


        
    Return 
    ------
    float, float 
         [normalisation, estimated error]
    """
    args = [param_profile[k] for k in ["stddev", "alpha", "amplitude_ratio", "theta", "ell", "xcentroid", "ycentroid"]]
    return integrate.dblquad(normalmoffat_profile,
                                 param_profile["xcentroid"]-xbounds, param_profile["xcentroid"]+xbounds,
                                 gfun = lambda x:  param_profile["ycentroid"]-ybounds,
                                 hfun = lambda x:  param_profile["ycentroid"]+ybounds,
                                 epsabs=epsabs, args=args)



# ========================= #
#                           #
#  Background               #
#                           #
# ========================= #
def tilted_plane(x, y,
                three_coefs):
    """ """
    return np.dot(np.asarray([np.ones(x.shape[0]), x, y]).T, three_coefs)


def curved_plane(x, y,
                five_coefs):
    """ """
    return np.dot(np.asarray([np.ones(x.shape[0]), x, y, x*y, x*x, y*y]).T, five_coefs)


# ========================= #
#                           #
#  Ellipticity              #
#                           #
# ========================= #
def get_elliptical_distance(x, y, xcentroid=0, ycentroid=0, ell=0, theta=0):
    """
    Parameters
    ----------
    x,y: [array]
        Cartesian Coordinates

    x0,y0: [float] -optional-
        Cartesian coordinate of the ellipse center
        
    ell: [float] -optional-
        Ellipticity [0<ell<1[
        
    theta: [float] -optional-
        Angle of the ellipse [radian]
    
    Returns
    -------
    array for float (elliptical distance)
    """
    c, s  = np.cos(theta), np.sin(theta)
    rot   = np.asarray([[c, s], [-s, c]])
    xx,yy = np.dot(rot, np.asarray([x-xcentroid, y-ycentroid]))
    return np.sqrt(xx**2 + (yy/(1-ell))**2)


################################
#                              #
#                              #
#     SLICE Model              #
#                              #
#                              #
################################

class _PSFSliceModel_( BaseModel ):
    """ Virtual PSFSlice Model Class. You need to define 
    - get_profile 
    - get_background 
    """
    PROFILE_PARAMETERS    = [] # TO BE DEFINED
    BACKGROUND_PARAMETERS = [] # TO BE DEFINED
    
    def __new__(cls,*arg,**kwarg):
        """ Black Magic allowing generalization of Polynomial models """
        # - Profile
        cls.FREEPARAMETERS     = list(cls.PROFILE_PARAMETERS)+list(cls.BACKGROUND_PARAMETERS)
        return super( _PSFSliceModel_, cls).__new__(cls)

    # ================= #
    #    Method         #
    # ================= #
    # ---------- #
    #  SETTER    #
    # ---------- #
    def setup(self, parameters):
        """ """
        self.param_profile    = {k:v for k,v in zip( self.PROFILE_PARAMETERS, parameters[:len(self.PROFILE_PARAMETERS)] )}
        self.param_background = {k:v for k,v in zip( self.BACKGROUND_PARAMETERS, parameters[len(self.PROFILE_PARAMETERS):] )} 
        
    # ---------- #
    #  GETTER    #
    # ---------- #
    def get_loglikelihood(self, x, y, z, dz):
        """ Measure the likelihood to find the data given the model's parameters.
        Set pdf to True to have the array prior sum of the logs (array not in log=pdf).
        In the Fitter define _get_model_args_() that should return the input of this
        """
        res = z - self.get_model(x, y)
        chi2 = np.nansum(res.flatten()**2/dz.flatten()**2)
        return -0.5 * ( chi2 - 2*self.get_logprior() )

    def get_logprior(self):
        """ If you need to return prior value. Do so here. """
        return 0
    
    def get_model(self, x, y):
        """ the profile + background model. """
        return self.get_profile(x,y) + self.get_background(x,y)

    # - To Be Defined
    def get_profile(self, x, y):
        """ The profile at the given positions """
        raise NotImplementedError("You must define the get_profile")
    
    def get_background(self, x, y):
        """ The background at the given positions """
        raise NotImplementedError("You must define the get_background")

# ======================================= #
#                                         #
#  BiNormal + Background    Models        #
#                                         #
# ======================================= #

class BiNormalFlat( _PSFSliceModel_ ):
    """ """
    NAME = "binormal-flat"
    PROFILE_PARAMETERS = ["amplitude",
                          "stddev", "stddev_ratio", "amplitude_ratio",
                          "theta", "ell",
                          "xcentroid", "ycentroid"]
    
    BACKGROUND_PARAMETERS = ["bkgd"]
    
    # ================== #
    #  Guess             #
    # ================== #
    def get_guesses(self, x, y, data,
                        xcentroid=None, xcentroid_err=2,
                        ycentroid=None, ycentroid_err=2):
        """ return a dictionary containing simple best guesses """
        flagok     = ~np.isnan(x*y*data)
        x          = x[flagok]
        y          = y[flagok]
        data       = data[flagok]
        
        ampl       = np.nanmax(data)
        if ycentroid is None or xcentroid is None:
            argmaxes   = np.argwhere(data>np.percentile(data,95)).flatten()

        if xcentroid is None:
            xcentroid  = np.nanmean(x[argmaxes])
        if ycentroid is None:
            ycentroid  = np.nanmean(y[argmaxes])
            
        background = np.percentile(data,10)
        low_bounds = -np.percentile(data,0.01)
        self._guess = dict( amplitude_guess=ampl * 5,
                            amplitude_boundaries= [None, None],
                            # - background
                            bkgd_guess=background, bkgd_boundaries=[None, np.percentile(data,99.9)],
                            # centroid
                            xcentroid_guess=xcentroid, xcentroid_boundaries=[xcentroid-xcentroid_err, xcentroid+xcentroid_err],
                            ycentroid_guess=ycentroid, ycentroid_boundaries=[ycentroid-ycentroid_err, ycentroid+ycentroid_err],
                            # ------------------------ #
                            # SEDM DEFAULT VARIABLES   #
                            # ------------------------ #
                            # Ellipticity
                            ell_guess=0.05, ell_boundaries=[0,0.4], ell_fixed=False,
                            theta_guess=1.5, theta_boundaries=[0,np.pi], theta_fixed=False,
                            # Size
                            stddev_guess = 1.3,
                            stddev_boundaries=[0.5, 5],
                            stddev_ratio_guess=2.,
                            stddev_ratio_boundaries=[1.1, 4],
                            stddev_ratio_fixed=False,
                            # Converges faster by allowing degenerated param...
                            # amplitude ratio
                            amplitude_ratio_guess = 3,
                            amplitude_ratio_fixed = False,
                            amplitude_ratio_boundaries = [1.5,5],
                           )
        return self._guess

    # ================== #
    #  Model             #
    # ================== #
    def get_profile(self, x, y):
        """ """
        return binormal_profile(x, y, **self.param_profile)
    
    def get_background(self,x,y):
        """ The background at the given positions """
        return self.param_background["bkgd"]

    def display_model(self, ax, rmodel, legend=True,
                          nobkgd=True,
                          cmodel = "C1",
                          cgaussian1 = "C0",cgaussian2 = "C2",
                          cbkgd="k", zorder=7, **kwargs):
        """ """
        # the decomposed binormal_profile
        n1 = _normal_(rmodel, scale=self.param_profile['stddev'])
        n2 = _normal_(rmodel, scale=self.param_profile['stddev']*self.param_profile['stddev_ratio'])

        coef1 = self.param_profile['amplitude_ratio']/(1.+self.param_profile['amplitude_ratio'])
        coef2 = 1./(1+self.param_profile['amplitude_ratio'])

        amplitude = self.param_profile['amplitude']
        # and its background
        background = 0 if nobkgd else self.param_background['bkgd']

        # - display background
        if not nobkgd:
            ax.axhline(background, ls=":",color=cbkgd, label="background",zorder=zorder)
        
        # - display details
        ax.plot(rmodel, background + n1*coef1*amplitude, ls="-.",color=cgaussian1, label="Core Gaussian",zorder=zorder,
                    **kwargs)
        ax.plot(rmodel, background + n2*coef2*amplitude, ls="-.",color=cgaussian2, label="Tail Gaussian",zorder=zorder,
                    **kwargs)
        # - display full model
        ax.plot(rmodel, background + (n2*coef2+n1*coef1)*amplitude, 
                    ls="-",color=cmodel,zorder=zorder+1, lw=2, label="PSF Model",
                    **kwargs)

        # - add the legend
        if legend:
            ax.legend(loc="best", ncol=1)
        
    # ============= #
    #  Properties   #
    # ============= #
    @property
    def centroid_guess(self):
        """ """
        return self._guess["xcentroid_guess"], self._guess["ycentroid_guess"]
    
    @property
    def centroid(self):
        """ """
        return self.fitvalues["xcentroid"], self.fitvalues["ycentroid"]
    
    
class BiNormalTilted( BiNormalFlat ):
    """ """
    NAME = "binormal-tilted"
    BACKGROUND_PARAMETERS = ["bkgd","bkgdx","bkgdy"]
    
    def get_background(self, x, y):
        """ The background at the given positions """
        return tilted_plane(x, y, [self.param_background[k] for k in self.BACKGROUND_PARAMETERS])
    
class BiNormalCurved( BiNormalFlat ):
    """ """
    NAME = "binormal-curved"
    BACKGROUND_PARAMETERS = ["bkgd","bkgdx","bkgdy","bkgdxy","bkgdxx","bkgdyy"]
    
    def get_background(self, x, y):
        """ The background at the given positions """
        return curved_plane(x, y, [self.param_background[k] for k in self.BACKGROUND_PARAMETERS])

# ======================================= #
#                                         #
#  NormalMoffat + Background  Models      #
#                                         #
# ======================================= #


class MoffatFlat( _PSFSliceModel_ ):
    """ """
    NAME = "moffat-flat"
    PROFILE_PARAMETERS = ["amplitude",
                          "alpha",  "beta",
                          "theta", "ell",
                          "xcentroid", "ycentroid"]
    
    BACKGROUND_PARAMETERS = ["bkgd"]
    
    # ================== #
    #  Guess             #
    # ================== #
    def get_guesses(self, x, y, data,
                        xcentroid=None, xcentroid_err=2,
                        ycentroid=None, ycentroid_err=2):
        """ return a dictionary containing simple best guesses """
        flagok     = ~np.isnan(x*y*data)
        x          = x[flagok]
        y          = y[flagok]
        data       = data[flagok]
        
        ampl       = np.nanmax(data)
        if ycentroid is None or xcentroid is None:
            argmaxes   = np.argwhere(data>np.percentile(data,95)).flatten()

        if xcentroid is None:
            xcentroid  = np.nanmean(x[argmaxes])
        if ycentroid is None:
            ycentroid  = np.nanmean(y[argmaxes])

        background = np.percentile(data,10)
        low_bounds = -np.percentile(data,0.01)
        self._guess = dict( amplitude_guess=ampl * 5,
                            amplitude_boundaries= [None, None],
                            # - background
                            bkgd_guess=background, bkgd_boundaries=[None, np.percentile(data,99.9)],

                            # centroid
                            xcentroid_guess=xcentroid, xcentroid_boundaries=[xcentroid-xcentroid_err, xcentroid+xcentroid_err],
                            ycentroid_guess=ycentroid, ycentroid_boundaries=[ycentroid-ycentroid_err, ycentroid+ycentroid_err],
                            # ------------------------ #
                            # SEDM DEFAULT VARIABLES   #
                            # ------------------------ #
                            # Ellipticity
                            ell_guess=0.05, ell_boundaries=[0,0.4], ell_fixed=False,
                            theta_guess=1.5, theta_boundaries=[0,np.pi], theta_fixed=False,
                            # Size
                            # moffat
                            alpha_guess=4.,
                            alpha_boundaries=[1., 8],
                            beta_guess=2.,
                            beta_boundaries=[0., None],
                            # Converges faster by allowing degenerated param...
                            # amplitude ratio
                           )
        return self._guess

    # ================== #
    #  Model             #
    # ================== #
    def get_logprior(self):
        """ If you need to return prior value. Do so here. """
        return 0

    def get_profile(self, x, y):
        """ """
        return moffat_profile(x, y, **self.param_profile)
    
    def get_background(self,x,y):
        """ The background at the given positions """
        return self.param_background["bkgd"]

    def display_model(self, ax, rmodel, legend=True,
                          nobkgd=True,
                          cmodel = "C1",
                          cgaussian1 = "C0",
                          cbkgd="k", zorder=7, **kwargs):
        """ """
        # the decomposed binormal_profile
        
        n1 = _moffat_(rmodel, alpha=self.param_profile['alpha'], beta=self.param_profile['beta'])

        amplitude = self.param_profile['amplitude']
        # and its background
        background = 0 if nobkgd else self.param_background['bkgd']

        # - display background
        if not nobkgd:
            ax.axhline(background, ls=":",color=cbkgd, label="background",zorder=zorder)
        
        # - display details
        ax.plot(rmodel, background + n1*amplitude, ls="-.",color=cgaussian1, label="Moffat", zorder=zorder,
                    **kwargs)
        
        # - display full model
        #ax.plot(rmodel, background + (n2*coef2+n1*coef1)*amplitude, 
        #            ls="-",color=cmodel,zorder=zorder+1, lw=2, label="PSF Model",
        #            **kwargs)

        # - add the legend
        if legend:
            ax.legend(loc="best", ncol=1)
        
    # ============= #
    #  Properties   #
    # ============= #
    @property
    def centroid_guess(self):
        """ """
        return self._guess["xcentroid_guess"], self._guess["ycentroid_guess"]
    
    @property
    def centroid(self):
        """ """
        return self.fitvalues["xcentroid"], self.fitvalues["ycentroid"]
    
class MoffatTilted( MoffatFlat ):
    """ """
    NAME = "binormal-tilted"
    BACKGROUND_PARAMETERS = ["bkgd","bkgdx","bkgdy"]
    
    def get_background(self, x, y):
        """ The background at the given positions """
        return tilted_plane(x, y, [self.param_background[k] for k in self.BACKGROUND_PARAMETERS])
    
class MoffatCurved( MoffatFlat ):
    """ """
    NAME = "binormal-curved"
    BACKGROUND_PARAMETERS = ["bkgd","bkgdx","bkgdy","bkgdxy","bkgdxx","bkgdyy"]
    
    def get_background(self, x, y):
        """ The background at the given positions """
        return curved_plane(x, y, [self.param_background[k] for k in self.BACKGROUND_PARAMETERS])

# ======================================= #
#                                         #
#  NormalMoffat + Background  Models      #
#                                         #
# ======================================= #


class NormalMoffatFlat( _PSFSliceModel_ ):
    """ """
    NAME = "normal/moffat-flat"
    PROFILE_PARAMETERS = ["amplitude",
                          "alpha", "amplitude_ratio",
                          "stddev", 
                          "theta", "ell",
                          "xcentroid", "ycentroid"]
    
    BACKGROUND_PARAMETERS = ["bkgd"]
    
    # ================== #
    #  Guess             #
    # ================== #
    def get_guesses(self, x, y, data,
                        xcentroid=None, xcentroid_err=2,
                        ycentroid=None, ycentroid_err=2):
        """ return a dictionary containing simple best guesses """
        flagok     = ~np.isnan(x*y*data)
        x          = x[flagok]
        y          = y[flagok]
        data       = data[flagok]
        
        ampl       = np.nanmax(data)
        if ycentroid is None or xcentroid is None:
            argmaxes   = np.argwhere(data>np.percentile(data,95)).flatten()

        if xcentroid is None:
            xcentroid  = np.nanmean(x[argmaxes])
        if ycentroid is None:
            ycentroid  = np.nanmean(y[argmaxes])

        background = np.percentile(data,10)
        low_bounds = -np.percentile(data,0.01)
        self._guess = dict( amplitude_guess=ampl * 5,
                            amplitude_boundaries= [None, None],
                            # - background
                            bkgd_guess=background, bkgd_boundaries=[None, np.percentile(data,99.9)],
                            # centroid
                            xcentroid_guess=xcentroid, xcentroid_boundaries=[xcentroid-xcentroid_err, xcentroid+xcentroid_err],
                            ycentroid_guess=ycentroid, ycentroid_boundaries=[ycentroid-ycentroid_err, ycentroid+ycentroid_err],
                            # ------------------------ #
                            # SEDM DEFAULT VARIABLES   #
                            # ------------------------ #
                            # Ellipticity
                            ell_guess=0.05, ell_boundaries=[0,0.4], ell_fixed=False,
                            theta_guess=1.5, theta_boundaries=[0,np.pi], theta_fixed=False,
                            # Size
                            stddev_guess = 1.3,
                            stddev_boundaries=[1., 2],
                            # moffat
                            alpha_guess=5.,
                            alpha_boundaries=[2., 10.],
                            #beta_guess=2.,
                            #beta_boundaries=[0., None],
                            # Converges faster by allowing degenerated param...
                            # amplitude ratio
                            amplitude_ratio_guess = 2,
                            amplitude_ratio_fixed = False,
                            amplitude_ratio_boundaries = [0.1,10],
                           )
        return self._guess

    # ================== #
    #  Model             #
    # ================== #
    def get_logprior(self):
        """ If you need to return prior value. Do so here. """
        return 0

    def get_profile(self, x, y):
        """ """
        return normalmoffat_profile(x, y, **self.param_profile)
    
    def get_background(self,x,y):
        """ The background at the given positions """
        return self.param_background["bkgd"]

    def display_model(self, ax, rmodel, legend=True,
                          nobkgd=True,
                          cmodel = "C1",
                          cgaussian1 = "C0",cgaussian2 = "C2",
                          cbkgd="k", zorder=7, **kwargs):
        """ """
        # the decomposed binormal_profile
        n1 = _normal_(rmodel,  scale=self.param_profile['stddev'])
        n2 = _moffat_(rmodel, alpha=self.param_profile['alpha'], beta=_alpha_to_beta_(self.param_profile['alpha']))

        coef1 = self.param_profile['amplitude_ratio']/(1.+self.param_profile['amplitude_ratio'])
        coef2 = 1./(1+self.param_profile['amplitude_ratio'])

        amplitude = self.param_profile['amplitude']
        # and its background
        background = 0 if nobkgd else self.param_background['bkgd']

        # - display background
        if not nobkgd:
            ax.axhline(background, ls=":",color=cbkgd, label="background",zorder=zorder)
        
        # - display details
        ax.plot(rmodel, background + n1*coef1*amplitude, ls="-.",color=cgaussian1, label="Core Gaussian",zorder=zorder,
                    **kwargs)
        ax.plot(rmodel, background + n2*coef2*amplitude, ls="-.",color=cgaussian2, label="Tail Moffat",zorder=zorder,
                    **kwargs)
        # - display full model
        ax.plot(rmodel, background + (n2*coef2+n1*coef1)*amplitude, 
                    ls="-",color=cmodel,zorder=zorder+1, lw=2, label="PSF Model",
                    **kwargs)

        # - add the legend
        if legend:
            ax.legend(loc="best", ncol=1)
        
    # ============= #
    #  Properties   #
    # ============= #
    @property
    def centroid_guess(self):
        """ """
        return self._guess["xcentroid_guess"], self._guess["ycentroid_guess"]
    
    @property
    def centroid(self):
        """ """
        return self.fitvalues["xcentroid"], self.fitvalues["ycentroid"]
    
class NormalMoffatTilted( NormalMoffatFlat ):
    """ """
    NAME = "normal/moffat-tilted"
    BACKGROUND_PARAMETERS = ["bkgd","bkgdx","bkgdy"]
    
    def get_background(self, x, y):
        """ The background at the given positions """
        return tilted_plane(x, y, [self.param_background[k] for k in self.BACKGROUND_PARAMETERS])
    
class NormalMoffatCurved( NormalMoffatFlat ):
    """ """
    NAME = "normal/moffat-curved"
    BACKGROUND_PARAMETERS = ["bkgd","bkgdx","bkgdy","bkgdxy","bkgdxx","bkgdyy"]
    
    def get_background(self, x, y):
        """ The background at the given positions """
        return curved_plane(x, y, [self.param_background[k] for k in self.BACKGROUND_PARAMETERS])



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from propobject import BaseObject
LBDAREF = 7000

# ========================= #
#                           #
#  Chromaticity             #
#                           #
# ========================= #
def stddev_chromaticity(lbda, stddevref, rho=-1/5., lbdaref=7000):
    """ Evolution of the standard deviation as a function of lbda. """
    return stddevref * (lbda / lbdaref)**(rho)

#############################
#                           #
#   Classes                 #
#                           #
#############################
class ADRModel( BaseObject ):
    """ """
    PROPERTIES = ["adr","unit","xref","yref"]

    def __init__(self,  adr, xref, yref, unit=1):
        """ """
        self.set_adr(adr, xref, yref, unit=unit)
        
    # =================== #
    #  Methods            #
    # =================== #
    def get_centroid(self, lbda):
        """ """
        return self.adr.refract(self.xref, self.yref, lbda, unit=self.unit)
    
    def set_adr(self, adr, xref, yref, unit=1):
        """ """
        self._properties["adr"]  = adr
        self._properties["xref"] = xref
        self._properties["yref"] = yref
        self._properties["unit"] = unit
        
    # =================== #
    #  Property           #
    # =================== #
    @property
    def adr(self):
        """ """
        return self._properties["adr"]

    @property
    def unit(self):
        """ """
        return self._properties["unit"]

    @property
    def xref(self):
        """ """
        return self._properties["xref"]
    
    @property
    def yref(self):
        """ """
        return self._properties["yref"]


class ChromaticNormalMoffat( BaseObject ):
    """ """
    PROPERTIES = ["stddev","alpha", "amplitude_ratio","lbda",
                      "adrmodel"]

    # ==================== #
    #  Methods             #
    # ==================== #
    def force_fit(self, cube, psfmodel="NormalMoffatCurved",
                      slice_width=None,
                      ell=None, theta=None, ellerr=None, thetaerr=None,
                      force_ellipse=True,
                      force_centroid=True,
                      force_stddev=True,force_alpha=True):
        """ 
        slice_width: [int/None]
            How may lbda-slices array do you combine.
            None=>1
            1 means, each slice individually
            2 means you stack 2 slices together (hence having a spectrum with only half the lbda-binning
            ...
        """
        from .fitter import fit_slice
        from astropy.stats import mad_std
        slpsf = {}
        # = Which wavelengths
        if slice_width is None or slice_width == 1:
            slice_width = None
            lbdas = cube.lbda
        elif slice_width<=0:
            raise ValueError("slice_width must be positive or None %d given"%slice_width)
        elif slice_width > 0:
            lbda_min,lbda_max = cube.lbda[::slice_width],cube.lbda[slice_width-1::slice_width]
            if len(lbda_min)>len(lbda_max):
                lbda_min = lbda_min[:len(lbda_max)]
            lbdas = np.mean([lbda_min,lbda_max], axis=0)

        stddev_lbdas = self.get_stddev(lbdas)

        # Moffat
        achromatic_moffat_alpha = self.get_moffat_alpha(False)
        amplitude_ratio = np.nanmean(self.amplitude_ratio)
        amplitude_ratioerr = mad_std(self.amplitude_ratio)*2
        profile_lbda = {"alpha_guess":achromatic_moffat_alpha[0],
                        "alpha_fixed": force_alpha,
                        "alpha_boundaries":[achromatic_moffat_alpha[0]-0.2,achromatic_moffat_alpha[0]+0.2],
                        "centroids_err":[0.05,0.05],
                        "force_centroid":force_centroid,
                        "stddev_fixed":force_stddev,
                        "amplitude_ratio_guess": np.nanmean(self.amplitude_ratio),
                        "amplitude_ratio_boundaries": [np.nanmax([0.1,amplitude_ratio-amplitude_ratioerr]), amplitude_ratio+amplitude_ratioerr]
                        }
        # Ellipse
        if ell is not None:
            profile_lbda["ell_guess"] = ell
            if ellerr is not None: profile_lbda["ell_boundaries"] = [ell-ellerr, ell+ellerr]
        if theta is not None:
            profile_lbda["theta_guess"] = theta
            if thetaerr is not None: profile_lbda["theta_boundaries"] = [theta-thetaerr, theta+thetaerr]
                
        if ell is not None and theta is not None and force_ellipse:
            profile_lbda["ell_fixed"]      = True
            profile_lbda["theta_fixed"]    = True
        # 
        # => All Slices
        #
        if slice_width is None:
            for i, l_ in enumerate(cube.lbda):
                profile_lbda["stddev_guess"]      = stddev_lbdas[i]
                profile_lbda["stddev_boundaries"] = [stddev_lbdas[i]-0.1,stddev_lbdas[i]+0.1]
                # Centroid
                profile_lbda["centroids"]      = self.get_position(l_)
                if i%40 == 0:
                    print(i,"/",len(cube.lbda))
                slpsf[i] = fit_slice(cube.get_slice(index=i, slice_object=True),
                                        psfmodel=psfmodel, **profile_lbda)
        # 
        # => Meta Slices
        #
        else:
            for i, l_ in  enumerate(zip(lbda_min,lbda_max)):
                profile_lbda["stddev_guess"]      = stddev_lbdas[i]
                profile_lbda["stddev_boundaries"] = [stddev_lbdas[i]-0.1,stddev_lbdas[i]+0.1]
                # Centroid
                profile_lbda["centroids"]      = self.get_position(np.mean(l_))
                if i%10 == 0:
                    print(i,"/",len(lbdas))
                slpsf[i] = fit_slice(cube.get_slice(lbda_min=l_[0],lbda_max=l_[1], slice_object=True),
                                        psfmodel=psfmodel, **profile_lbda)
        return slpsf
    
    
    # -------- #
    #  GETTER  #
    # -------- #
    def get_position(self, lbda):
        """ """
        return self.adrmodel.get_centroid(lbda)
    
    def get_stddev(self, lbda, lbdaref=LBDAREF):
        """ """
        stddevref, rho = self.fit_stddev_parameters()
        return stddev_chromaticity(lbda, stddevref, rho=rho, lbdaref=lbdaref)
    
    def get_moffat_alpha(self, weighted_mean=True):
        """ """
        if weighted_mean and self.alpha_err is not None:
            flagin = ~np.isnan(self.alpha*self.alpha_err)
            return np.average(self.alpha[flagin], weights=1/self.alpha_err[flagin]**2), np.nanstd(self.alpha)/ np.sqrt(self.ndata)
        
        return np.nanmean(self.alpha), np.nanstd(self.alpha)/ np.sqrt(self.ndata)
    
    # -------- #
    #  SETTER  #
    # -------- #
    def set_adrmodel(self, adrmodel):
        """ """
        self._properties["adrmodel"] = adrmodel

        
    def set_data(self, stddev, alpha, amplitude_ratio, lbda,
                     stddev_err=None, alpha_err=None, amplitude_ratio_err=None):
        """ """
        self._properties["stddev"]              = stddev
        self._properties["alpha"]               = alpha
        self._properties["amplitude_ratio"]     = amplitude_ratio
        self._properties["stddev_err"]          = stddev_err
        self._properties["alpha_err"]           = alpha_err
        self._properties["amplitude_ratio_err"] = amplitude_ratio_err
        
        self._properties["lbda"]            = lbda

    # --------- #
    #  FITTER   #
    # --------- #
    def fit_stddev_parameters(self, adjust_errors=True, intrinsic=0):
        """ """
        from scipy.optimize import minimize
        stddev_err = self.stddev_err if self.stddev_err is not None else 1.
        if intrinsic>0: stddev_err = np.sqrt(stddev_err**2 + intrinsic**2)
        
        def _fmin_(param): 
            return np.nansum( np.sqrt((self.stddev-stddev_chromaticity(self.lbda, *param))**2/stddev_err**2))

        res  = minimize(_fmin_, [np.nanmedian(self.stddev), -1/5.], bounds=[[0.5,10], [-1,1]], options={"disp":0})
        if self.stddev_err is not None:
            chi2_dof = res["fun"] / len(self.stddev-2)
            if chi2_dof>3 and adjust_errors:
                print("Adjusting error")
                from pysedm.utils.tools import fit_intrinsic
                intrinsic = fit_intrinsic(self.stddev, stddev_chromaticity(self.lbda, *res["x"]), self.stddev_err, len(self.stddev-2), intrinsic_guess=None)
                return self.fit_stddev_parameters( adjust_errors=False, intrinsic = intrinsic/1.4)
        
        return res["x"]
    
    # ==================== #
    #  Properties          #
    # ==================== #
    
    # - PSF Properties
    @property
    def stddev(self):
        """ """
        return self._properties["stddev"]

    @property
    def alpha(self):
        """ """
        return self._properties["alpha"]

    @property
    def amplitude_ratio(self):
        """ """
        return self._properties["amplitude_ratio"]

    @property
    def stddev_err(self):
        """ """
        return self._properties["stddev_err"]

    @property
    def alpha_err(self):
        """ """
        return self._properties["alpha_err"]

    @property
    def amplitude_ratio_err(self):
        """ """
        return self._properties["amplitude_ratio_err"]
    
    
    @property
    def lbda(self):
        """ """
        return self._properties["lbda"]

    @property
    def ndata(self):
        """ """
        return len(self.lbda)

    @property
    def adrmodel(self):
        """ """
        return self._properties["adrmodel"]




""" Generic simple tools. """

import numpy as np

__all__ = ["kwargs_update"]


def kwargs_update(default,**kwargs):
    """
    """
    k = default.copy()
    for key,val in kwargs.items():
        k[key] = val
        
    return k



def fit_intrinsic(data, model, errors, dof, intrinsic_guess=None):
    """ Get the most optimal intrinsic dispersion given the current fitted standardization parameters. 
    
    The optimal intrinsic magnitude dispersion is the value that has to be added in quadrature to 
    the errors such that the chi2/dof is 1.
    Returns
    -------
    float (intrinsic dispersion)
    """
    from scipy.optimize import fmin
    def get_intrinsic_chi2dof(intrinsic):
        return np.abs( np.nansum((data-model)**2/(errors**2+intrinsic**2)) / dof - 1)
    
    if intrinsic_guess is None:
        intrinsic_guess = np.nanmedian(errors)
        
    return fmin(get_intrinsic_chi2dof, intrinsic_guess, disp=0)[0]
